#  Copyright 2022-present, the Waterdip Labs Pvt. Ltd.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
from abc import ABC, abstractmethod
from dataclasses import asdict
from enum import Enum
from typing import Dict, List

from datachecks.core.configuration.configuration import MetricConfiguration as MetricConfigurationV1
from datachecks.core.datasource.data_source import DataSource, SearchIndexDataSource, DataSourceManager, SQLDatasource


class MetricsType(str, Enum):
    ROW_COUNT = 'row_count'
    DOCUMENT_COUNT = 'document_count'


class MetricIdentity:

    @staticmethod
    def generate_identity(
            metric_type: MetricsType,
            metric_name: str,
            data_source: DataSource = None,
            index_name: str = None,
            table_name: str = None,
            row_name: str = None
    ):
        identifiers = []
        if data_source:
            identifiers.append(data_source.data_source_name)
        identifiers.append(metric_type.value)
        identifiers.append(metric_name)
        if index_name:
            identifiers.append(index_name)
        if table_name:
            identifiers.append(table_name)
        if row_name:
            identifiers.append(row_name)

        return ".".join([str(p) for p in identifiers])


class Metric(ABC):
    """
    Metric is a class that represents a metric that is generated by a data source.
    """
    def __init__(
            self,
            name: str,
            data_source: DataSource,
            metric_type: MetricsType,
            filter: Dict = None
    ):
        self.name: str = name
        self.data_source = data_source
        self.metric_type = metric_type
        self.data_filter = filter
        self.metric_identity = MetricIdentity.generate_identity(
            metric_type=metric_type,
            metric_name=name,
            data_source=data_source
        )

    @abstractmethod
    def _generate_metric_value(self):
        pass

    def get_value(self):
        value = {
                "identity": self.metric_identity,
                "metricName": self.name,
                "metricType": self.metric_type.value,
                "value": self._generate_metric_value(),
                "dataSourceName": self.data_source.data_source_name,
                "@timestamp": datetime.datetime.utcnow().isoformat()
            }
        if "index_name" in self.__dict__:
            value["index_name"] = self.__dict__["index_name"]
        elif "table_name" in self.__dict__:
            value["table_name"] = self.__dict__["table_name"]

        if "row_name" in self.__dict__:
            value["row_name"] = self.__dict__["row_name"]

        return value


class RowMetrics(Metric, ABC):

    def __init__(
            self,
            name: str,
            data_source: DataSource,
            table_name: str,
            row_name: str,
            metric_type=MetricsType,
            filter: Dict = None
    ):
        super().__init__(
            name=name,
            data_source=data_source,
            metric_type=metric_type,
            filter=filter
        )
        self.table_name = table_name
        self.row_name = row_name

        self.metric_identity = MetricIdentity.generate_identity(
            metric_type=MetricsType.DOCUMENT_COUNT,
            metric_name=name,
            data_source=data_source,
            table_name=table_name,
            row_name=row_name
        )

    @property
    def get_row_name(self):
        return self.row_name


class DocumentCountMetrics(Metric):

    def __init__(
            self,
            name: str,
            data_source: DataSource,
            index_name: str,
            filter: Dict = None
    ):
        super().__init__(
            name=name,
            data_source=data_source,
            metric_type=MetricsType.DOCUMENT_COUNT,
            filter=filter
        )
        self.index_name = index_name
        self.metric_identity = MetricIdentity.generate_identity(
            metric_type=MetricsType.DOCUMENT_COUNT,
            metric_name=name,
            data_source=data_source,
            index_name=index_name
        )

    def validate_data_source(self):
        return isinstance(self.data_source, SearchIndexDataSource)

    def _generate_metric_value(self):
        if isinstance(self.data_source, SearchIndexDataSource):
            return self.data_source.query_get_document_count(
                index_name=self.index_name,
                filter=self.data_filter['search_query']
            )
        else:
            raise ValueError("Invalid data source type")


class RowCountMetrics(Metric):

    def __init__(
        self,
        name: str,
        data_source: DataSource,
        table_name: str,
        filter: Dict = None
    ):
        super().__init__(
            name=name,
            data_source=data_source,
            metric_type=MetricsType.ROW_COUNT,
            filter=filter
        )
        self.table_name = table_name
        self.metric_identity = MetricIdentity.generate_identity(
            metric_type=MetricsType.ROW_COUNT,
            metric_name=name,
            data_source=data_source,
            table_name=table_name
        )

    def validate_data_source(self):
        return isinstance(self.data_source, SQLDatasource)

    def _generate_metric_value(self):
        if isinstance(self.data_source, SQLDatasource):
            return self.data_source.query_get_row_count(
                table=self.table_name,
                filter=self.data_filter['sql_query']
            )
        else:
            raise ValueError("Invalid data source type")


class MetricManager:

    def __init__(
            self,
            metric_config: Dict[str, List[MetricConfigurationV1]],
            data_source_manager: DataSourceManager
    ):
        self.data_source_manager = data_source_manager
        self.metrics = {}
        self._build_metrics(config=metric_config)

    def _build_metrics(self, config: Dict[str, List[MetricConfigurationV1]]):
        for data_source, metric_list in config.items():
            for metric_config in metric_list:
                if metric_config.metric_type == MetricsType.DOCUMENT_COUNT:

                    metric = DocumentCountMetrics(
                        name=metric_config.name,
                        data_source=self.data_source_manager.get_data_source(data_source),
                        filter=asdict(metric_config.filter),
                        index_name=metric_config.index
                    )
                    self.metrics[metric.metric_identity] = metric
                elif metric_config.metric_type == MetricsType.ROW_COUNT:
                    metric = RowCountMetrics(
                        name=metric_config.name,
                        data_source=self.data_source_manager.get_data_source(data_source),
                        filter=asdict(metric_config.filter),
                        table_name=metric_config.table
                    )
                    self.metrics[metric.metric_identity] = metric
                else:
                    raise ValueError("Invalid metric type")

    @property
    def get_metrics(self):
        return self.metrics

    def get_metric(self, metric_identity: str):
        return self.metrics.get(metric_identity, None)
